clc; clear; close all; set(0,'DefaultFigureVisible','on')

% What files are we importing in?
dataFile = 'HynneSmooth.txt';

% Define our noisy dataset
metabList = 1:22;
numResampleIters = 250;
numRandIters = 50;
CoV = .15;
numReplicates = 5;
numSamples = 61; % 0 to 30m, at 0.5m intervals
timeUnit = 'min';
concUnit = 'mM';


% Re-Order of metabolites from the COPASI output
columnList = 2:23;

%	2	GlcX	 - 	Extracellular glucose
%	3	Glc	     - 	Cytosolic glucose
%	4	G6P	     - 	Glucose-6-Phosphate
%	5	F6P	     - 	Fructose-6-Phosphate
%	6	FBP	     - 	Fructose 1,6-bisphosphate
%	7	DHAP	 - 	Dihydroxyacetone phosphate
%	8	GAP	     - 	Glyceraldehyde 3-phosphate
%	9	BPG	     - 	1,3-Bisphosphoglycerate
%	10	PEP	     - 	Phosphoenolpyruvate
%	11	Pyr	     - 	Pyruvate
%	12	ACA	     - 	Acetaldehyde
%	13	ACAX	 - 	Extracellular acetaldehyde
%	14	EtOH	 - 	EtOH
%	15	EtOHX	 - 	Extracellular ethanol
%	16	Glyc	 - 	Glycerol
%	17	GlycX	 - 	Extracellular glycerol
%	18	CNX	     - 	Extracellular cyanide
%	19	AMP	     - 	AMP
%	20	ADP	     - 	ADP
%	21	ATP	     - 	ATP
%	22	NAD	     - 	NAD
%	23	NADH	 - 	NADH


% Load in our data
dataStruct = importdata(dataFile);
    trueDataAll = dataStruct.data(:,columnList);
    smoothTime = dataStruct.data(:,1);
    colNames = dataStruct.colheaders(columnList)';
    
% Take the derivative of the true, smooth data (Finite Difference Approx)
trueDataDerivAll = (trueDataAll(3:end,:) - trueDataAll(1:end-2,:))./repmat((smoothTime(3:end) - smoothTime(1:end-2)),1,size(trueDataAll,2));
derivSmoothTime = smoothTime(2:end-1);

% Get the data at sampling frequency
sampleTime = linspace(smoothTime(1),smoothTime(end),numSamples)';
[~, idxSample] =      min(abs(repmat(smoothTime,1,size(sampleTime,1))-repmat(sampleTime',size(smoothTime,1),1)));
[~, idxSampleDeriv] = min(abs(repmat(derivSmoothTime,1,size(sampleTime,1))-repmat(sampleTime',size(derivSmoothTime,1),1)));
noiselessDataAll = trueDataAll(idxSample,:);
noiselessDataDerivAll = trueDataDerivAll(idxSampleDeriv,:);


for metab_j = metabList;

    metabNum = metab_j;
    
    % Base underlying data
    trueData = repmat(trueDataAll(:,metab_j),1,numRandIters);
    trueDataDeriv = repmat(trueDataDerivAll(:,metab_j),1,numRandIters);
    noiselessData = repmat(noiselessDataAll(:,metab_j),1,numRandIters);
    noiselessDataDeriv = repmat(noiselessDataDerivAll(:,metab_j),1,numRandIters);
 
    
    % Generate the noisy datasets
    baseNoisyData = zeros(size(noiselessDataAll,1),numRandIters);
    
    for iteration_k = 1:numRandIters

        % Reinitialize storage space...
        noisyDataGen = zeros(size(noiselessDataAll,1),numReplicates);

        % Generate the noisy measurements
        for j=1:numRandIters
            for i = 1:size(noiselessDataAll,1)

                for k = 1:numReplicates
                    % Add noise to the original measurements
                    noisyDataGen(i,j,k) = CoV*noiselessDataAll(i,metab_j)*randn(1,1) + noiselessDataAll(i,metab_j);
                end

                % Get the average of the noisy data
               baseNoisyData(i,j) = mean(noisyDataGen(i,j,:));

            end
        end

    end

    % For each noisy dataset, now we need to generate the resampled sets 
    noisyDataMatrix = repmat({zeros(size(baseNoisyData))},numResampleIters,1);

    for iteration_k = 1:numResampleIters

        % Reinitialize storage space...
        noisyData = zeros(size(baseNoisyData,1),size(baseNoisyData,2),numReplicates);

        % Generate the noisy measurements
        for j=1:size(baseNoisyData,2)
            for i = 1:size(baseNoisyData,1)

                for k = 1:numReplicates
                    % Add noise to the original measurements
                    noisyData(i,j,k) = CoV*baseNoisyData(i,j)*randn(1,1) + baseNoisyData(i,j);
                end

                % Get the average of the noisy data
                noisyDataMatrix{iteration_k}(i,j) = mean(noisyData(i,j,:));

            end
        end
        
    end
    
    
    save(sprintf('hynneResampleNoisy_metab%02d.mat',metab_j),...
        'metabNum',...
        'CoV',...
        'numReplicates',...
        'numRandIters',...
        'numResampleIters',...
        'timeUnit',...
        'concUnit',...
        'colNames',...
        'trueData',...
        'trueDataDeriv',...
        'noiselessData',...
        'noiselessDataDeriv',...
        'baseNoisyData',...
        'noisyDataMatrix',...
        'sampleTime',...
        'smoothTime',...
        'derivSmoothTime',...
        'idxSampleDeriv')

end
