clc; clear; close all; set(0,'DefaultFigureVisible','on')

% Specify what we're looking for right here up top
baseDatasetFileName = '../hynneData/hynneNoisy.mat';
fileNameStem = '../hynneFit/output_hynneDirect/directFitParams';
metabIndexList = [1:16, 18:22];


ticStart = tic;

fprintf('Generating Direct Fit Method plots from %s\n\n',baseDatasetFileName);

fullFuncList = {'poly2';
                'poly3';
                'poly4';
                'poly5';
                'rat11';
                'rat22';
                'rat31';
                'impls'};
            
% Pull the master dataset
baseDataset = load(baseDatasetFileName);
metabNameList = baseDataset.colNames;
noiseCoV = baseDataset.CoV;

% Get time arrays
sampleTime = baseDataset.sampleTime;
smoothTime = baseDataset.smoothTime;
smoothTimeDeriv = baseDataset.derivSmoothTime;
hdTime = linspace(sampleTime(1),sampleTime(end),10*(length(sampleTime)-1)+1)';
timeUnit = baseDataset.timeUnit;

% Get our true data arrays
[~, idxHdConc] = min(abs(repmat(smoothTime,1,size(hdTime,1))-repmat(hdTime',size(smoothTime,1),1)));
[~, idxHdDeriv] = min(abs(repmat(smoothTimeDeriv,1,size(hdTime,1))-repmat(hdTime',size(smoothTimeDeriv,1),1)));

trueConcSmooth = baseDataset.trueData;
trueDerivSmooth = baseDataset.trueDataDeriv;
trueConcSample = baseDataset.noiselessData;
trueDerivSample = baseDataset.trueDataDeriv(baseDataset.idxSampleDeriv,:);
trueConcHd = baseDataset.trueData(idxHdConc,:);
trueDerivHd = baseDataset.trueDataDeriv(idxHdDeriv,:);
concUnit = baseDataset.concUnit;

% Final clean-up/set-up
clear baseDataset

set_k = 1;


% Loop through our functions
for func_k = fullFuncList'

    % Which function are we checking right now?
    fprintf('Reading Function: %s\n',func_k{1});
    
    % Let's run through out list of metabs
    for metab_k = metabIndexList

        % More robust handling for when cases screw up
        try

            % Get the parameters for this particlar function & metabolite
            load(sprintf('%s_%s_metab%02d.mat',fileNameStem,func_k{1},metab_k));

            % Record info about which function and metabolite we're
            % looking at, for future reference
            currentFunc{set_k} = func_k{1};
            currentMetabNum(set_k) = metab_k;
            currentMetabName{set_k} = metabNameList{metab_k};
            fprintf('\tMetab %02d - %s\n',metab_k,currentMetabName{set_k});

            % Now let's get cracking
            directConcSample = zeros(length(sampleTime),numIters);
            directConcHd = zeros(length(hdTime),numIters);
            directConcSmooth = zeros(length(smoothTime),numIters);
            directDerivSample = zeros(length(sampleTime),numIters);
            directDerivHd = zeros(length(hdTime),numIters);
            directDerivSmooth = zeros(length(smoothTimeDeriv),numIters);

            % We're calculated the parameters elsewhere:
            % Regenerate time courses for the direct fit
            for noisySet_k = 1:numIters
                directConcSample(:,noisySet_k) = evalSmoothingFcn(directParamSolution(:,noisySet_k),sampleTime,currentFunc{set_k});
                directConcHd(:,noisySet_k) = evalSmoothingFcn(directParamSolution(:,noisySet_k),hdTime,currentFunc{set_k});
                directConcSmooth(:,noisySet_k) = evalSmoothingFcn(directParamSolution(:,noisySet_k),smoothTime,currentFunc{set_k});
                directDerivSample(:,noisySet_k) = evalSmoothingDeriv(directParamSolution(:,noisySet_k),sampleTime,currentFunc{set_k});
                directDerivHd(:,noisySet_k) = evalSmoothingDeriv(directParamSolution(:,noisySet_k),hdTime,currentFunc{set_k});
                directDerivSmooth(:,noisySet_k) = evalSmoothingDeriv(directParamSolution(:,noisySet_k),smoothTimeDeriv,currentFunc{set_k});
            end


            % We have everything we need to plot trajectories!
            try
                plotTrajectoriesDirect
            end


            % Some scaling and penalty factors for our metrics
            denominator = length(sampleTime)-nnz(~isnan(directParamSolution(:,1)));
            hdDenom = length(hdTime) * ((length(sampleTime)-nnz(~isnan(directParamSolution(:,1))))/length(sampleTime));
            
            concScaleSample = sqrt(mean((trueConcSample(:,metab_k)).^2));
            derivScaleSample = sqrt(mean((trueDerivSample(:,metab_k)).^2));
            concScaleHd = sqrt(mean((trueConcHd(:,metab_k)).^2));
            derivScaleHd = sqrt(mean((trueDerivHd(:,metab_k)).^2));
            
            % Error metrics
            directConcAccHd{set_k} = sqrt(sum((directConcHd - repmat(trueConcHd(:,metab_k),1,numIters)).^2,1))/(hdDenom*concScaleHd);
            directDerivAccHd{set_k} = sqrt(sum((directDerivHd - repmat(trueDerivHd(:,metab_k),1,numIters)).^2,1))/(hdDenom*derivScaleHd);

            % Some of the plot methods are interested in the medians of the metrics
            medianDirectConcAccHd(set_k) = median(directConcAccHd{set_k},2);
            medianDirectDerivAccHd(set_k) = median(directDerivAccHd{set_k},2);
            
            % How consistent are my fits? --> Robustness of performance
            directConcDispersion(set_k) = median(sqrt(sum(((directConcHd - repmat(median(directConcHd,2),1,numIters)).^2)/hdDenom,1))/concScaleHd);
            directDerivDispersion(set_k) = median(sqrt(sum(((directDerivHd - repmat(median(directDerivHd,2),1,numIters)).^2)/hdDenom,1))/derivScaleHd);
            directConcRobustness(set_k) = mean(sqrt(sum(((directConcHd - repmat(median(directConcHd,2),1,numIters)).^2)/hdDenom,1))/concScaleHd);
            directDerivRobustness(set_k) = mean(sqrt(sum(((directDerivHd - repmat(median(directDerivHd,2),1,numIters)).^2)/hdDenom,1))/derivScaleHd);
            
            
            % Housecleaning in between calculations
            clear('directConcHd',...
                  'directConcSample',...
                  'directConcSmooth',...
                  'directDerivHd',...
                  'directDerivSample',...
                  'directDerivSmooth',...
                  'denominator',...
                  'hdDenom');
            
            % FINALLY - if we've made it got here, everything is hunky
            % dory, and we can save what we calculated for this set!
            set_k = set_k + 1;

        catch
            fprintf('\tWarning: Error during analysis of ''%s_%s_metab%02d.mat''\n',fileNameStem,func_k{1},metab_k);
        end

    end
end

save testMatDirect

 % Generatebar plots - compare across functions
try
    plotBarGraphsDirect
end

% Generatescatter plots - compare across functions
try
    plotScatterGraphsDirect
end

% Generate Rank Bar Plots and Table - compare across functions
try
    calcRankMetaMetricsDirect
end

fprintf('\nAnalysis Complete.\n')
toc(ticStart)
fprintf('\n')
